__all__ = [
    'typednamedtuple',
    'Heap',
    'TimeBasedHeap',
    'OmniHashableMixin',
    'Singleton',
    'DictObject'
]

from .singleton import Singleton
from .structures import *
from .typednamedtuple import *
from .dict_object import *



import types
import typing as tp
from abc import abstractmethod

__all__ = [
    'BaseExceptionHandler',
    'exception_handler',
    'ALWAYS_FIRST',
    'ALWAYS_LAST',
    'NORMAL_PRIORITY'
]

ALWAYS_FIRST = -1000
NORMAL_PRIORITY = 0
ALWAYS_LAST = 1000


class BaseExceptionHandler:
    def __init__(self, priority=NORMAL_PRIORITY):
        self.priority = priority

    def install(self):
        from .global_eh import GlobalExcepthook
        GlobalExcepthook().add_hook(self)
        return self

    def uninstall(self):
        from .global_eh import GlobalExcepthook
        GlobalExcepthook().remove_hook(self)

    @abstractmethod
    def handle_exception(self, type_: tp.Optional[type], value,
                         traceback: types.TracebackType) -> tp.Optional[bool]:
        """Return True to intercept the exception. It won't be propagated to other handlers."""
        pass


class FunctionExceptionHandler(BaseExceptionHandler):
    def __init__(self, fun: tp.Callable, priority: int = NORMAL_PRIORITY):
        super(FunctionExceptionHandler, self).__init__(priority)
        self.fun = fun

    def handle_exception(self, type_, value, traceback):
        if type_ == SystemExit:
            return
        return self.fun(type_, value, traceback)


def exception_handler(priority: int = NORMAL_PRIORITY):
    """
    Convert a callable to an FunctionExceptionHandler. Usage

        @exception_handler(priority=-10)
        def handle_exc(type, val, traceback):
            ...

    :return: ExceptionHandler instance
    """

    if not isinstance(priority, int):
        raise TypeError('Did you forget to use it as @exception_handler() ?')

    def outer(fun):
        return FunctionExceptionHandler(fun, priority=priority)

    return outer



import logging
import typing as tp

logger = logging.getLogger(__name__)


class DictObject(dict):
    def __getattr__(self, item: str) -> tp.Any:
        try:
            return self[item]
        except KeyError as e:
            raise AttributeError(repr(e))

    def __setattr__(self, key: str, value: tp.Any) -> None:
        self[key] = value

    def __delattr__(self, key: str) -> None:
        try:
            del self[key]
        except KeyError as e:
            raise AttributeError(repr(e))



# coding=UTF-8
"""
Just useful objects to make your coding nicer every day
"""
from __future__ import print_function, absolute_import, division

from .algos import merge_dicts
from .concurrent import Monitor, RMonitor, CallableGroup, LockedDataset
from .recast_exceptions import rethrow_as, silence_excs
from .structures import TimeBasedHeap, Heap, typednamedtuple, OmniHashableMixin, Singleton, \
    DictObject
from .decorators import precondition, for_argument
from .fun_static import static_var

__all__ = [
    'typednamedtuple', 'OmniHashableMixin'
                       'TimeBasedHeap', 'Heap', 'CallableGroup', 'DictObject',
    'Monitor', 'RMonitor', 'CallableGroup', 'LockedDataset', 'merge_dicts',
    'for_argument',
    'precondition', 'PreconditionError',
    'rethrow_as', 'silence_excs',
    'Singleton',
    'static_var'
]
